﻿using System;
using System.Collections.Generic;

using PT.Scheduler;
using PT.Scheduler.Simulation.Customizations;
using PT.Transmissions;

namespace PlanetTogether.Samples.CIP
{
    /// <summary>
    /// 
    /// </summary>
    public class EndOfSimulation : EndOfSimulationCustomization, ICustomizationRequiresInitialization
    {
        public override string Name => "CIP Processing";

        public override string Description => "Part of the CIP Package.";
        
        private CipRequiredCapacity m_requiredCapacityCust = null;
        private Dictionary<BaseId, TimeSpan> m_activityPostProcessing = new Dictionary<BaseId, TimeSpan>();

        private bool m_sendingTimeAdjustment;

        private void BroadcastTimeAdjustment(ScenarioDetail a_sd, ScenarioBaseT a_t)
        {
            //A time adjustment is the fastest simulation. Use the touchT to trigger a time adjustment.
            //Since we are not re-sequencing operations, this will work to update the post processing from the RequiredCapacity customization
            ScenarioTouchT touchT = new ScenarioTouchT();
            touchT.sequenced = false;
            touchT.TransmissionSender = PTTransmissionBase.TransmissionSenderType.PTSystem;
            AddTransmissionsToProcess(1, touchT);
        }


        protected override void SimulationInitialization(ScenarioDetail a_sd, ScenarioDetail.SimulationType a_simulationType, ScenarioBaseT a_transmission)
        {
            Utilities.InitializeUDFs(a_sd);
            
            //Remove previous extra transmissions
            CleanTransmissionsProcess();
        }

        private void DetermineActivityPostProcessing(Resource a_res)
        {
            if (!Utilities.IsCipEnabled(a_res)) return;

            Utilities.ECipType cipType = Utilities.GetCipType(a_res);
            switch (cipType)
            {
                case Utilities.ECipType.Duration:
                    AddCipDurationProcessing(a_res);
                    break;
                case Utilities.ECipType.Qty:
                    AddCipQtyProcessing(a_res);
                    break;
            }
        }

        private void AddCipDurationProcessing(Resource a_res)
        {
            ResourceBlockList.Node prev = a_res.Blocks.First;
            TimeSpan durationRun = Utilities.GetHrsRunBeforeClock(a_res);
            TimeSpan maxRunDurationBeforeCip = Utilities.GetCipHrs(a_res);
            TimeSpan cipDuration = Utilities.GetCipDuration(a_res);
            while(prev != null)
            {
                InternalActivity prevActivity = prev.Data.Batch.FirstActivity;
                durationRun += prevActivity.ScheduledRunSpan;
                ResourceBlockList.Node next = prev.Next;
                if (next != null)
                {
                    InternalActivity nextAct = next.Data.Batch.FirstActivity;

                    if (durationRun + nextAct.ScheduledRunSpan > maxRunDurationBeforeCip)
                    {
                        //CIP needed here
                        m_activityPostProcessing.Add(prevActivity.Id, cipDuration);
                        durationRun = TimeSpan.Zero;
                    }
                }

                prev = next;
            }
        }

        private void AddCipQtyProcessing(Resource a_res)
        {
            ResourceBlockList.Node prev = a_res.Blocks.First; //Start at the first scheduled activity
            //Get our initial qty since last CIP
            decimal qtyProduced = Utilities.GetQtyProducedBeforeClock(a_res);
            //Get the max that can be produced before we need another CIP
            decimal maxQtyBeforeCip = Utilities.GetCipQty(a_res);
            //Get how long the CIP should run for
            TimeSpan cipDuration = Utilities.GetCipDuration(a_res);
            //Go through each scheduled activity in order to calculate required CIPs
            while(prev != null)
            {
                InternalActivity prevActivity = prev.Data.Batch.FirstActivity;
                qtyProduced += prevActivity.RemainingQty; //Add the qty required for this activity
                ResourceBlockList.Node next = prev.Next;
                if (next != null)
                {
                    InternalActivity nextAct = next.Data.Batch.FirstActivity;

                    //Check if the next activity is over our max qty before a CIP
                    if (qtyProduced + nextAct.RemainingQty > maxQtyBeforeCip)
                    {
                        //CIP needed here. Store this activity and the CIP duration to be used in the RequiredCapacity customization
                        m_activityPostProcessing.Add(prevActivity.Id, cipDuration);
                        qtyProduced = 0;
                    }
                }

                prev = next;
            }
        }

        private void DetermineActivityPostProcessing(ScenarioDetail a_sd)
        {
            m_activityPostProcessing.Clear();

            foreach(Plant p in a_sd.PlantManager)
            {
                foreach(Department dept in p.Departments)
                {
                    foreach (Resource res in dept.Resources)
                    {
                        DetermineActivityPostProcessing(res);
                    }
                }
            }
        }

        protected override void EndOfSimulationCleanup(ScenarioDetail.SimulationType a_simType, ScenarioBaseT a_t, ScenarioDetail a_sd)
        {
            //Every action will cause another time adjustment to be sent
            if (!m_sendingTimeAdjustment)
            {
                m_sendingTimeAdjustment = true;
                DetermineActivityPostProcessing(a_sd);
                //Trigger a second simulation here so that the RequiredCapacity customization can use calculated values to add the CIP post processing
                BroadcastTimeAdjustment(a_sd, a_t);
            }
            else
            {
                m_sendingTimeAdjustment = false;
                m_activityPostProcessing.Clear();
            }
        }

        //Part of ICustomizationRequiresInitialization
        //These two functions can be used to synchronize different customization points.
        public void InitializeAfterCreation(ScenarioDetail a_sd)
        {
        }

        /// <summary>
        /// Pass the CIP data over to the required capacity customization
        /// </summary>
        /// <param name="a_sd"></param>
        public void RestoreReferences(ScenarioDetail a_sd)
        {
            if (a_sd.CustomizationInstances.SchedulerCustomization is CipRequiredCapacity capacityCust)
            {
                m_requiredCapacityCust = capacityCust;
                m_requiredCapacityCust.SetActivityPostProcessing(m_activityPostProcessing);
            }
        }
    }
}
