﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using PT.Scheduler;

namespace PlanetTogether.Samples.CIP
{
    public class Schedulability : PT.Scheduler.Simulation.Customizations.SchedulabilityCustomization
    {
        public override string Name => "CIP Scheduling";

        public override string Description => "Part of the CIP Package.";

        /**
         * returns the last activity scheduled on the resource or null if 
         * no activities are scheduled on the resource.
         */
        private InternalActivity GetLastScheduledActivity(Resource a_res)
        {
            return a_res.Blocks.Last?.Data.Batch.FirstActivity;
        }

        protected override bool IsSchedulable(ScenarioDetail a_sd, ScenarioDetail.SimulationType a_simType, long a_clock, long a_simulationClock, BaseResource a_res, BaseActivity a_activity, SchedulableInfo a_schedulableInfo, out long o_postEventTime)
        {
            o_postEventTime = -1;

            // only apply this add-in logic during optimize.
            //if (a_simType == ScenarioDetail.SimulationType.TimeAdjustment || a_simType == ScenarioDetail.SimulationType.ConstraintsChangeAdjustment)
            //{
            //    return true;
            //}

            InternalActivity act = a_activity as InternalActivity;
            Resource res = a_res as Resource;

            InternalActivity lastScheduledAct = GetLastScheduledActivity(res);
            if (lastScheduledAct == null)
            {
                // this is the first activity being scheduled on the resource. allow it to schedule.
                return true;
            }

            if (!Utilities.IsCipEnabled(res))
            {
                return true;
            }

            switch (Utilities.GetCipType(res))
            {
                case Utilities.ECipType.Duration:
                    return VerifyDurationCip(res, act, a_simulationClock, out o_postEventTime);
                case Utilities.ECipType.Qty:
                    return VerifyQtyCip(res, act, a_simulationClock, out o_postEventTime);
            }

            return true;
        }

        private bool VerifyDurationCip(Resource a_res, InternalActivity a_act, long a_simulationClock, out long o_postEventTime)
        {
            o_postEventTime = -1;
            //Qty
            ResourceBlockList.Node prev = a_res.Blocks.First; //Start at the first scheduled activity
            TimeSpan durationRun = Utilities.GetHrsRunBeforeClock(a_res);
            TimeSpan maxRunDurationBeforeCip = Utilities.GetCipHrs(a_res);
            TimeSpan cipDuration = Utilities.GetCipDuration(a_res);
            while(prev != null)
            {
                InternalActivity prevActivity = prev.Data.Batch.FirstActivity;
                durationRun += prevActivity.ScheduledRunSpan;
                ResourceBlockList.Node next = prev.Next;
                if (next != null)
                {
                    InternalActivity nextAct = next.Data.Batch.FirstActivity;

                    if (durationRun + nextAct.ScheduledRunSpan > maxRunDurationBeforeCip)
                    {
                        //CIP needed here. Reset duration
                        durationRun = TimeSpan.Zero;
                    }
                }

                prev = next;
            }

            if (durationRun + a_act.ScheduledRunSpan > maxRunDurationBeforeCip)
            {
                //A CIP is needed before this activity. It must be delayed
                // calculate available capacity between the last scheduled act end date and 
                // a_simulationClock. a_simulationClock is the datetime that act is attempting to schedule to start at.
                InternalActivity lastScheduledAct = GetLastScheduledActivity(a_res);
                TimeSpan available = a_res.ResourceCapacityIntervals.GetTotalSchedulableCapacityBetweenDates(lastScheduledAct.ScheduledEndDate, new DateTime(a_simulationClock));
                if (available >= cipDuration)
                {
                    return true;
                }

                // we can use o_postEventTime to suggest to the schedule engine the next time the activity should attempt to schedule.
                // in our case, we'll set this to a_simulationClock + how much more time we need to create the gap required.
                o_postEventTime = a_simulationClock + (cipDuration - available).Ticks;
                return false;
            }

            return true;
        }

        private bool VerifyQtyCip(Resource a_res, InternalActivity a_act, long a_simulationClock, out long o_postEventTime)
        {
            o_postEventTime = -1;

            if (a_res.Name == "Huettlin")
            {
                if (a_act.Operation.ManufacturingOrder.Job.Name == "MRP -2147467505")
                {
                    bool b = false;
                }
            }

            //Qty. This calculation is very similar to the EndOfSimulation calculation, but is not quite the same. 
            ResourceBlockList.Node prev = a_res.Blocks.First;//Start at the first scheduled activity
            //Get our initial qty since last CIP
            decimal qtyProduced = Utilities.GetQtyProducedBeforeClock(a_res);
            //Get the max that can be produced before we need another CIP
            decimal maxQtyBeforeCip = Utilities.GetCipQty(a_res);
            //Get how long the CIP should run for
            TimeSpan cipDuration = Utilities.GetCipDuration(a_res);
            //Go through each scheduled activity in order to calculate required CIPs
            while(prev != null)
            {
                InternalActivity prevActivity = prev.Data.Batch.FirstActivity;
                qtyProduced += prevActivity.RemainingQty; //Add the qty required for this activity
                ResourceBlockList.Node next = prev.Next;
                if (next != null)
                {
                    InternalActivity nextAct = next.Data.Batch.FirstActivity;

                    //Check if the next activity is over our max qty before a CIP
                    if (qtyProduced + nextAct.RemainingQty > maxQtyBeforeCip)
                    {
                        //CIP needed here. Reset qtyProduced
                        qtyProduced = 0;
                    }
                }

                prev = next;
            }

            //Now check if the activity that is trying to schedule will cause a CIP
            if (qtyProduced + a_act.RemainingQty > maxQtyBeforeCip)
            {
                //A CIP is needed before this activity. It must be delayed
                // calculate available capacity between the last scheduled act end date and 
                // a_simulationClock. a_simulationClock is the datetime that act is attempting to schedule to start at.
                InternalActivity lastScheduledAct = GetLastScheduledActivity(a_res);
                TimeSpan available = a_res.ResourceCapacityIntervals.GetTotalSchedulableCapacityBetweenDates(lastScheduledAct.ScheduledEndOfRunDate, new DateTime(a_simulationClock));
                if (available >= cipDuration)
                {
                    //We are past the end of the CIP, this activity can schedule here.
                    return true;
                }

                // we can use o_postEventTime to suggest to the schedule engine the next time the activity should attempt to schedule.
                // in our case, we'll set this to a_simulationClock + how much more time we need to create the gap required.
                o_postEventTime = a_simulationClock + (cipDuration - available).Ticks;
                return false;
            }

            return true;
        }
    }
}
