﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using PT.Scheduler;
using PT.SchedulerDefinitions;

namespace PlanetTogether.Samples.CIP
{
    internal class Utilities
    {
        internal static string EnableCipUdf = "Enable CIP";
        internal static string CipAfterQtyUdf = "CIP After Quantity";
        internal static string CipAfterRunHrsUdf = "CIP After Run Hours";
        internal static string CipDurationMinutesUdf = "CIP Duration Minutes";
        internal static string QtyProducedSinceLastCipUdf = "Qty Since Last CIP";
        internal static string HrsRunSinceLastCipUdf = "Hrs Run Since Last CIP";
        internal enum ECipType { Duration, Qty, Invalid }

        internal static void InitializeUDFs(ScenarioDetail a_sd)
        {
            foreach (Resource resource in a_sd.PlantManager.GetResourceList())
            {
                if (resource.UserFields == null)
                {
                    resource.UserFields = new UserFieldList();
                }

                if (resource.UserFields.Find(CipAfterQtyUdf) == null)
                {
                    resource.UserFields.Add(new UserField(CipAfterQtyUdf, 0m, UserField.UDFTypes.Decimal));
                }

                if (resource.UserFields.Find(CipAfterRunHrsUdf) == null)
                {
                    resource.UserFields.Add(new UserField(CipAfterRunHrsUdf, 0d, UserField.UDFTypes.Double));
                }

                if (resource.UserFields.Find(QtyProducedSinceLastCipUdf) == null)
                {
                    resource.UserFields.Add(new UserField(QtyProducedSinceLastCipUdf, 0m, UserField.UDFTypes.Decimal));
                } 
                
                if (resource.UserFields.Find(HrsRunSinceLastCipUdf) == null)
                {
                    resource.UserFields.Add(new UserField(HrsRunSinceLastCipUdf, 0d, UserField.UDFTypes.Double));
                }

                if (resource.UserFields.Find(CipDurationMinutesUdf) == null)
                {
                    resource.UserFields.Add(new UserField(CipDurationMinutesUdf, 0d, UserField.UDFTypes.Double));
                } 

                if (resource.UserFields.Find(EnableCipUdf) == null)
                {
                    resource.UserFields.Add(new UserField(EnableCipUdf, false, UserField.UDFTypes.Boolean));
                }
            }
        }

        /// <summary>
        /// Check the resource UDFs to determine if CIP should be managed on this resource
        /// </summary>
        internal static bool IsCipEnabled(BaseResource a_res)
        {
            if (a_res.UserFields?.Find(EnableCipUdf)?.DataValue is bool enabled && enabled)
            {
                var cipType = GetCipType(a_res);
                if (cipType != ECipType.Invalid)
                {
                    return true;
                }
            }

            return false;
        }

        /// <summary>
        /// Check the resource UDF to determine how long a CIP runs on this resource.
        /// </summary>
        internal static TimeSpan GetCipDuration(BaseResource a_res)
        {
            if (a_res.UserFields?.Find(CipDurationMinutesUdf)?.DataValue is double cipDurationMinutes)
            {
                return TimeSpan.FromMinutes(cipDurationMinutes);
            }

            return TimeSpan.Zero;
        }

        /// <summary>
        /// Check the resource UDF to determine the max qty that can be run before a CIP is needed.
        /// </summary>
        internal static decimal GetCipQty(BaseResource a_res)
        {
            if (a_res.UserFields?.Find(CipAfterQtyUdf)?.DataValue is decimal cipQty)
            {
                return cipQty;
            }

            return 0m;
        }

        internal static TimeSpan GetCipHrs(BaseResource a_res)
        {
            if (a_res.UserFields?.Find(CipAfterRunHrsUdf)?.DataValue is double cipDurationHrs)
            {
                return TimeSpan.FromHours(cipDurationHrs);
            }

            return TimeSpan.Zero;
        }
        
        public static ECipType GetCipType(BaseResource a_res)
        {
            decimal cipQty = GetCipQty(a_res);
            TimeSpan cipDuration = GetCipHrs(a_res);
            if (cipQty != 0m && cipDuration != TimeSpan.Zero)
            {
                //Both are defined, this is invalid
                return ECipType.Invalid;
            }

            if (cipQty == 0m && cipDuration == TimeSpan.Zero)
            {
                //Neither are defined
                return ECipType.Invalid;
            }

            if (cipQty != 0m)
            {
                return ECipType.Qty;
            }
            
            return ECipType.Duration;
        }


        /// <summary>
        /// Check the resource UDF to determine the imported value for how much Qty has been run since the last CIP
        /// </summary>
        public static decimal GetQtyProducedBeforeClock(BaseResource a_res)
        {
            if (a_res.UserFields?.Find(QtyProducedSinceLastCipUdf)?.DataValue is decimal cipQty)
            {
                return cipQty;
            }

            return 0m;
        }
        
        /// <summary>
        /// Check the resource UDF to determine the imported value for how much duration has been run since the last CIP
        /// </summary>
        public static TimeSpan GetHrsRunBeforeClock(BaseResource a_res)
        {
            if (a_res.UserFields?.Find(HrsRunSinceLastCipUdf)?.DataValue is double hours)
            {
                return TimeSpan.FromHours(hours);
            }

            return TimeSpan.Zero;
        }
    }
}
